"""
Types for the spotdl package.
"""

from spotdl.types.song import Song
from spotdl.types.playlist import Playlist
from spotdl.types.album import Album
from spotdl.types.artist import Artist
from spotdl.types.saved import Saved



"""
YTMusic module for downloading and searching songs.
"""

from typing import Any, Dict, List, Optional

from ytmusicapi import YTMusic
from slugify import slugify
from rapidfuzz import fuzz

from spotdl.providers.audio.base import AudioProvider
from spotdl.types import Song
from spotdl.utils.formatter import (
    create_song_title,
    parse_duration,
    create_search_query,
)


class YouTubeMusic(AudioProvider):
    """
    YouTube Music audio provider class
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the YouTube Music API

        ### Arguments
        - args: Arguments passed to the `AudioProvider` class.
        - kwargs: Keyword arguments passed to the `AudioProvider` class.
        """

        super().__init__(*args, **kwargs)
        self.client = YTMusic()

    def search(self, song: Song) -> Optional[str]:
        """
        Search for a song on YouTube Music.

        ### Arguments
        - song: The song to search for.

        ### Returns
        - The url of the best match or None if no match was found.
        """

        if self.search_query:
            search_query = create_search_query(
                song, self.search_query, False, None, True
            )
        else:
            # search for song using isrc if it's available
            if song.isrc:
                isrc_results = self.get_results(song.isrc, filter="songs")

                if len(isrc_results) == 1:
                    isrc_result = self.order_results([isrc_results[0]], song)
                    if len(isrc_result) == 1:
                        isrc_link, isrc_score = isrc_result.popitem()

                        if isrc_score > 90:
                            # print(f"# RETURN URL - {isrc_link} - isrc score")
                            return isrc_link

            search_query = create_song_title(song.name, song.artists).lower()

        # Query YTM by songs only first, this way if we get correct result on the first try
        # we don't have to make another request
        song_results = self.get_results(search_query, filter="songs")

        if self.filter_results:
            # Order results
            songs = self.order_results(song_results, song)
        else:
            songs = {}
            if len(song_results) > 0:
                songs = {song_results[0]["link"]: 100}

        # song type results are always more accurate than video type,
        # so if we get score of 80 or above
        # we are almost 100% sure that this is the correct link
        if len(songs) != 0:
            # get the result with highest score
            best_result = max(songs, key=lambda k: songs[k])

            if songs[best_result] >= 80:
                # print(f"# RETURN URL - {best_result} - song >= 80")
                return best_result

        # We didn't find the correct song on the first try so now we get video type results
        # add them to song_results, and get the result with highest score
        video_results = self.get_results(search_query, filter="videos")

        if self.filter_results:
            # Order video results
            videos = self.order_results(video_results, song)
        else:
            videos = {}
            if len(video_results) > 0:
                videos = {video_results[0]["link"]: 100}

        # Merge songs and video results
        results = {**songs, **videos}

        # No matches found
        if not results:
            return None

        result_items = list(results.items())

        # Sort results by highest score
        sorted_results = sorted(result_items, key=lambda x: x[1], reverse=True)

        # print(f"# RETURN URL - {sorted_results[0][0]} - sorted")

        # Get the result with highest score
        # and return the link
        return sorted_results[0][0]

    def get_results(self, search_term: str, **kwargs) -> List[Dict[str, Any]]:
        """
        Get results from YouTube Music API and simplify them

        ### Arguments
        - search_term: The search term to search for.
        - kwargs: other keyword arguments passed to the `YTMusic.search` method.

        ### Returns
        - A list of simplified results (dicts)
        """

        results = self.client.search(search_term, **kwargs)

        # Simplify results
        simplified_results = []
        for result in results:
            if result is None or result.get("videoId") is None:
                continue

            simplified_results.append(
                {
                    "name": result["title"],
                    "type": result["resultType"],
                    "link": f"https://youtube.com/watch?v={result['videoId']}",
                    "album": result.get("album", {}).get("name")
                    if result.get("album")
                    else None,
                    "duration": parse_duration(result.get("duration")),
                    "artists": ", ".join(map(lambda a: a["name"], result["artists"])),
                }
            )

        return simplified_results

    def order_results(
        self, results: List[Dict[str, Any]], song: Song
    ) -> Dict[str, Any]:
        """
        Filter results based on the song's metadata.

        ### Arguments
        - results: The results to filter.
        - song: The song to filter by.

        ### Returns
        - A dict of filtered results.
        """

        # Slugify some variables
        slug_song_name = slugify(song.name)
        slug_album_name = slugify(song.album_name)
        slug_song_artist = slugify(song.artist)
        slug_song_title = slugify(
            create_song_title(song.name, song.artists)
            if not self.search_query
            else create_search_query(song, self.search_query, False, None, True)
        )

        # Assign an overall avg match value to each result
        links_with_match_value = {}
        for result in results:
            # Slugify result title
            slug_result_name = slugify(result["name"])
            slug_result_artists = slugify(result["artists"])
            slug_result_album = (
                slugify(result["album"]) if result.get("album") else None
            )

            # check for common words in result name
            sentence_words = slug_song_name.split("-")
            common_word = any(
                word != "" and word in slug_result_name for word in sentence_words
            )

            # DEBUG CODE
            # print(f"#############################")
            # print(f"slug_result_name: {slug_result_name}")
            # print(f"slug_result_artists: {slug_result_artists}")
            # print(f"slug_result_album: {slug_result_album}")
            # print(f"slug_song_name: {slug_song_name}")
            # print(f"slug_album_name: {slug_album_name}")
            # print(f"slug_song_artist: {slug_song_artist}")
            # print(f"slug_song_title: {slug_song_title}")
            # print(f"URL - {result['link']}")
            # print("-----------------------------")

            # skip results that have no common words in their name
            # print(f"common_word: {common_word}")
            if not common_word:
                continue

            artist_match_number = 0
            for artist in song.artists:
                # print("song.artist", artist)
                # print("slugify artist", slugify(artist))
                for slugified_result in [slug_result_name, slug_result_artists]:
                    artist_match_number += (
                        1 if slugify(artist) in slugified_result else 0
                    )

            artist_match = artist_match_number * 100 / len(song.artists)
            # print("first artist_match: ", artist_match)

            # If we didn't find any artist match,
            # we fallback to channel name match
            if artist_match <= 50 and result["type"] != "song":
                channel_name_match = fuzz.partial_token_sort_ratio(
                    slugify(song.artist),
                    slug_result_artists,
                )

                if channel_name_match > artist_match_number:
                    artist_match = channel_name_match
                    # print("second artist_match: ", artist_match)

            # skip results with artist match lower than 70%
            if artist_match < 70:
                # print("! artist_match < 70 - skipping")
                continue

            # Calculate name match
            # for different result types
            if result["type"] == "song":
                name_match = fuzz.partial_token_sort_ratio(
                    slug_result_name,
                    slug_song_name,
                )
            else:
                # We are almost certain that this result
                # contains the correct song artist
                # so if the title doesn't contain the song artist in it
                # we append slug_song_artist to the title
                if artist_match > 90 and slug_song_artist not in slug_result_name:
                    name_match = fuzz.partial_token_sort_ratio(
                        f"{slug_song_artist}-{slug_result_name}",
                        slug_song_title,
                    )
                else:
                    name_match = fuzz.partial_token_sort_ratio(
                        slug_result_name,
                        slug_song_title,
                    )

            # Drop results with name match lower than 50%
            # print(f"name_match: {name_match}")
            if name_match < 50:
                # print("! name_match < 50 - skipping")
                continue

            # Find album match
            album_match = 0.0

            # Calculate album match only for songs
            if result["type"] == "song":
                if slug_result_album:
                    album_match = fuzz.partial_ratio(slug_result_album, slug_album_name)

            # Calculate time match
            delta = result["duration"] - song.duration
            non_match_value = (delta**2) / song.duration * 100
            time_match = 100 - non_match_value

            # Calculate total match
            average_match = (artist_match + name_match + time_match) / 3

            # print(f"album_match: {album_match}")
            # print(f"time_match: {time_match}")
            # print(f"average_match: {average_match}")

            if (
                result["type"] == "song"
                and slug_result_album
                and fuzz.partial_ratio(
                    slug_album_name, slug_result_name, score_cutoff=95
                )
                and slug_result_album == slug_album_name
            ):
                # If the result album name is similar to the song album name
                # and the result album name is the same as the song album name
                # we add album match to the average match
                average_match = (
                    artist_match + album_match + name_match + time_match
                ) / 4

            # the results along with the avg Match
            links_with_match_value[result["link"]] = average_match

        return links_with_match_value



"""
Artist module for retrieving artist data from Spotify.
"""

from dataclasses import dataclass
from typing import Any, Dict, List
from spotdl.types.song import SongList
from spotdl.utils.spotify import SpotifyClient
from spotdl.types.song import Song


class AlbumError(Exception):
    """
    Base class for all exceptions related to albums.
    """


@dataclass(frozen=True)
class Album(SongList):
    """
    Album class for retrieving album data from Spotify.
    """

    artist: Dict[str, Any]

    @classmethod
    def from_url(cls, url: str) -> "Album":
        """
        Parse an album from a Spotify URL.

        ### Arguments
        - url: The URL of the album.

        ### Returns
        - The Album object.
        """

        metadata = Album.get_metadata(url)

        urls = cls.get_urls(url)

        # Remove songs without id (country restricted/local tracks)
        # And create song object for each track
        songs: List[Song] = [Song.from_url(url) for url in urls]

        return cls(
            **metadata,
            songs=songs,
            urls=urls,
        )

    @staticmethod
    def get_urls(url: str) -> List[str]:
        """
        Get urls for all songs in album.

        ### Arguments
        - url: The URL of the album.

        ### Returns
        - A list of urls.
        """

        spotify_client = SpotifyClient()

        album_response = spotify_client.album_tracks(url)
        if album_response is None:
            raise AlbumError(
                "Couldn't get metadata, check if you have passed correct album id"
            )

        tracks = album_response["items"]

        # Get all tracks from album
        while album_response["next"]:
            album_response = spotify_client.next(album_response)

            # Failed to get response, break the loop
            if album_response is None:
                break

            tracks.extend(album_response["items"])

        if album_response is None:
            raise AlbumError(f"Failed to get album response: {url}")

        return [
            track["external_urls"]["spotify"]
            for track in tracks
            if track and track.get("id")
        ]

    @staticmethod
    def get_metadata(url: str) -> Dict[str, Any]:
        """
        Get metadata for album.

        ### Arguments
        - url: The URL of the album.

        ### Returns
        - A dictionary with metadata.
        """

        spotify_client = SpotifyClient()

        album_metadata = spotify_client.album(url)
        if album_metadata is None:
            raise AlbumError(
                "Couldn't get metadata, check if you have passed correct album id"
            )

        return {
            "name": album_metadata["name"],
            "artist": album_metadata["artists"][0],
            "url": url,
        }



"""
MusixMatch lyrics provider.
"""

from typing import List, Optional
from urllib.parse import quote

import requests

from bs4 import BeautifulSoup

from spotdl.providers.lyrics.base import LyricsProvider


class MusixMatch(LyricsProvider):
    """
    MusixMatch lyrics provider class.
    """

    def get_lyrics(self, name: str, artists: List[str], **kwargs) -> Optional[str]:
        """
        Try to get lyrics from musixmatch

        ### Arguments
        - name: The name of the song.
        - artists: The artists of the song.
        - kwargs: Additional arguments.

        ### Returns
        - The lyrics of the song or None if no lyrics were found.
        """

        try:
            track_search = kwargs.get("track_search", False)

            artists_str = ", ".join(
                artist for artist in artists if artist.lower() not in name.lower()
            )

            # quote the query so that it's safe to use in a url
            # e.g "Au/Ra" -> "Au%2FRa"
            query = quote(f"{name} - {artists_str}", safe="")

            # search the `tracks page` if track_search is True
            if track_search:
                query += "/tracks"

            search_url = f"https://www.musixmatch.com/search/{query}"
            search_resp = requests.get(search_url, headers=self.headers)

            search_soup = BeautifulSoup(search_resp.text, "html.parser")
            song_url_tag = search_soup.select_one("a[href^='/lyrics/']")

            # song_url_tag being None means no results were found on the
            # All Results page, therefore, we use `track_search` to
            # search the tracks page.
            if song_url_tag is None:
                # track_serach being True means we are already searching the tracks page.
                if track_search:
                    return None

                lyrics = self.get_lyrics(name, artists, track_search=True)
                return lyrics

            song_url = "https://www.musixmatch.com" + str(song_url_tag.get("href", ""))
            lyrics_resp = requests.get(song_url, headers=self.headers)

            lyrics_soup = BeautifulSoup(lyrics_resp.text, "html.parser")
            lyrics_paragraphs = lyrics_soup.select("p.mxm-lyrics__content")
            lyrics = "\n".join(i.get_text() for i in lyrics_paragraphs)

            return lyrics
        except Exception:
            return None



"""
Save module for the console.
"""

import json
import concurrent.futures

from typing import List, Optional
from pathlib import Path

from spotdl.download.downloader import Downloader
from spotdl.utils.search import parse_query
from spotdl.utils.m3u import create_m3u_file


def save(
    query: List[str],
    downloader: Downloader,
    save_path: Path,
    m3u_file: Optional[str] = None,
    preload: bool = False,
) -> None:
    """
    Save metadata from spotify to the disk.

    ### Arguments
    - query: list of strings to search for.
    - downloader: Already initialized downloader instance.
    - save_path: Path to save the songs to.
    - m3u_file: Path to the m3u file to save the songs to.
    - preload: If True, preload the songs.

    ### Notes
    - This function is multi-threaded.
    """

    # Parse the query
    songs = parse_query(query, downloader.threads)
    save_data = [song.json for song in songs]

    if preload:
        save_data = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=downloader.threads
        ) as executor:
            future_to_song = {
                executor.submit(downloader.search, song): song for song in songs
            }
            for future in concurrent.futures.as_completed(future_to_song):
                song = future_to_song[future]
                try:
                    data, _ = future.result()
                    if data is None:
                        downloader.progress_handler.error(
                            f"Could not find a match for {song.display_name}"
                        )
                        continue

                    downloader.progress_handler.log(
                        f"Found url for {song.display_name}: {data}"
                    )
                    save_data.append({**song.json, "download_url": data})
                except Exception as exc:
                    downloader.progress_handler.error(
                        f"{song} generated an exception: {exc}"
                    )

    # Save the songs to a file
    with open(save_path, "w", encoding="utf-8") as save_file:
        json.dump(save_data, save_file, indent=4, ensure_ascii=False)

    if m3u_file:
        create_m3u_file(
            m3u_file, songs, downloader.output, downloader.output_format, False
        )

    downloader.progress_handler.log(
        f"Saved {len(save_data)} song{'s' if len(save_data) > 1 else ''} to {save_path}"
    )



"""
Playlist module for retrieving playlist data from Spotify.
"""

from dataclasses import dataclass
from typing import Any, Dict, List

from spotdl.types.song import SongList
from spotdl.utils.spotify import SpotifyClient
from spotdl.types.song import Song


class PlaylistError(Exception):
    """
    Base class for all exceptions related to playlists.
    """


@dataclass(frozen=True)
class Playlist(SongList):
    """
    Playlist class for retrieving playlist data from Spotify.
    """

    description: str
    author_url: str
    author_name: str

    @classmethod
    def from_url(cls, url: str) -> "Playlist":
        """
        Load playlist info and tracks from a Spotify playlist URL.

        ### Arguments
        - url: The URL of the playlist.

        ### Returns
        - The Playlist object.
        """

        metadata = Playlist.get_metadata(url)

        # Get urls
        urls = cls.get_urls(url)

        # Remove songs without id (country restricted/local tracks)
        # And create song object for each track
        tracks = [Song.from_url(url) for url in urls]

        return cls(
            **metadata,
            songs=tracks,
            urls=urls,
        )

    @staticmethod
    def get_urls(url: str) -> List[str]:
        """
        Get URLs of all tracks in a playlist.

        ### Arguments
        - url: The URL of the playlist.

        ### Returns
        - A list of urls.
        """

        spotify_client = SpotifyClient()
        tracks = []

        playlist_response = spotify_client.playlist_items(url)
        if playlist_response is None:
            raise PlaylistError(f"Wrong playlist id: {url}")

        tracks = playlist_response["items"]

        # Get all tracks from playlist
        while playlist_response["next"]:
            playlist_response = spotify_client.next(playlist_response)

            # Failed to get response, break the loop
            if playlist_response is None:
                break

            # Add tracks to the list
            tracks.extend(playlist_response["items"])

        return [
            track["track"]["external_urls"]["spotify"]
            for track in tracks
            if track is not None
            and track.get("track") is not None
            and track.get("track").get("id")
        ]

    @staticmethod
    def get_metadata(url: str) -> Dict[str, Any]:
        """
        Get metadata for a playlist.

        ### Arguments
        - url: The URL of the playlist.

        ### Returns
        - A dictionary with metadata.
        """

        spotify_client = SpotifyClient()

        playlist = spotify_client.playlist(url)
        if playlist is None:
            raise PlaylistError("Invalid playlist URL.")

        return {
            "name": playlist["name"],
            "url": url,
            "description": playlist["description"],
            "author_url": playlist["external_urls"]["spotify"],
            "author_name": playlist["owner"]["display_name"],
        }



"""
Module for formatting songs into strings.
Contains functions to create search queries and song titles
and file names.
"""

import re
import warnings

from typing import List, Optional
from pathlib import Path
from slugify import slugify
from yt_dlp.utils import sanitize_filename

from spotdl.types import Song

VARS = [
    "{title}",
    "{artists}",
    "{artist}",
    "{album}",
    "{album-artist}",
    "{genre}",
    "{disc-number}",
    "{disc-count}",
    "{duration}",
    "{year}",
    "{original-date}",
    "{track-number}",
    "{tracks-count}",
    "{isrc}",
    "{track-id}",
    "{publisher}",
    "{list-length}",
    "{list-position}",
    "{list-name}",
    "{output-ext}",
]


def create_song_title(song_name: str, song_artists: List[str]) -> str:
    """
    Create the song title.

    ### Arguments
    - song_name: the name of the song
    - song_artists: the list of artists of the song

    ### Returns
    - the song title

    ### Notes
    - Example: "Artist1, Artist2 - Song Name"

    """

    joined_artists = ", ".join(song_artists)
    if len(song_artists) >= 1:
        return f"{joined_artists} - {song_name}"

    return song_name


def sanitize_string(string: str) -> str:
    """
    Sanitize the filename to be used in the file system.

    ### Arguments
    - string: the string to sanitize

    ### Returns
    - the sanitized string
    """

    output = string

    # this is windows specific (disallowed chars)
    output = "".join(char for char in output if char not in "/?\\*|<>")

    # double quotes (") and semi-colons (:) are also disallowed characters but we would
    # like to retain their equivalents, so they aren't removed in the prior loop
    output = output.replace('"', "'").replace(":", "-")

    return output


def format_query(
    song: Song,
    template: str,
    santitize: bool,
    file_extension: Optional[str] = None,
    short: bool = False,
) -> str:
    """
    Replace template variables with the actual values.

    ### Arguments
    - song: the song object
    - template: the template string
    - santitize: whether to sanitize the string
    - file_extension: the file extension to use
    - short: whether to use the short version of the template

    ### Returns
    - the formatted string
    """

    if "{output-ext}" in template and file_extension is None:
        raise ValueError("file_extension is None, but template contains {output-ext}")

    if (
        any(k in template for k in ["{list-length}", "{list-position}", "{list-name}"])
        and song.song_list is None
    ):
        # If the template contains {list-length} or {list-position} or {list-name},
        # but the song_list is None, replace them with empty strings
        for k in ["{list-length}", "{list-position}", "{list-name}"]:
            template = template.replace(k, "")
            template = template.replace(r"//", r"/")

    # If template has only {output-ext}, fix it
    # This can happen if the template consits of only list values
    # and song.song_list is None
    if template in ["/.{output-ext}", ".{output-ext}"]:
        template = "{artists} - {title}.{output-ext}"

    # Remove artists from the list that are already in the title
    artists = [
        artist for artist in song.artists if slugify(artist) not in slugify(song.name)
    ]

    # Add the main artist again to the list
    if len(artists) == 0 or artists[0] != song.artists[0]:
        artists.insert(0, song.artists[0])

    artists_str = ", ".join(artists)

    # the code below is valid, song_list is actually checked for None
    formats = {
        "{title}": song.name,
        "{artists}": song.artists[0] if short is True else artists_str,
        "{artist}": song.artists[0],
        "{album}": song.album_name,
        "{album-artist}": song.album_artist,
        "{genre}": song.genres[0] if len(song.genres) > 0 else "",
        "{disc-number}": song.disc_number,
        "{disc-count}": song.disc_count,
        "{duration}": song.duration,
        "{year}": song.year,
        "{original-date}": song.date,
        "{track-number}": song.track_number,
        "{tracks-count}": song.tracks_count,
        "{isrc}": song.isrc,
        "{track-id}": song.song_id,
        "{publisher}": song.publisher,
        "{output-ext}": file_extension,
    }

    if song.song_list and any(
        k in template for k in ["{list-length}", "{list-position}", "{list-name}"]
    ):
        try:
            index = song.song_list.songs.index(song)
        except ValueError:
            index = song.song_list.urls.index(song.url)

        formats.update(
            {
                "{list-name}": song.song_list.name,  # type: ignore
                "{list-position}": str(index + 1).zfill(
                    len(str(song.song_list.length))
                ),
                "{list-length}": song.song_list.length,
            }
        )

    if santitize:
        # sanitize the values in formats dict
        for key, value in formats.items():
            if value is None:
                continue

            formats[key] = sanitize_string(str(value))

    # Replace all the keys with the values
    for key, value in formats.items():
        template = template.replace(key, str(value))

    return template


def create_search_query(
    song: Song,
    template: str,
    santitize: bool,
    file_extension: Optional[str] = None,
    short: bool = False,
) -> str:
    """
    Create the search query for the song.

    ### Arguments
    - song: the song object
    - template: the template string
    - santitize: whether to sanitize the string
    - file_extension: the file extension to use
    - short: whether to use the short version of the template

    ### Returns
    - the formatted string
    """

    # If template does not contain any of the keys,
    # append {artist} - {title} at the beggining of the template
    if not any(key in template for key in VARS):
        template = "{artist} - {title}" + template

    return format_query(song, template, santitize, file_extension, short=short)


def create_file_name(
    song: Song,
    template: str,
    file_extension: str,
    short: bool = False,
) -> Path:
    """
    Create the file name for the song, by replacing template variables with the actual values.

    ### Arguments
    - song: the song object
    - template: the template string
    - file_extension: the file extension to use
    - short: whether to use the short version of the template

    ### Returns
    - the formatted string as a Path object
    """

    # If template does not contain any of the keys,
    # append {artists} - {title}.{output-ext} to it
    if not any(key in template for key in VARS):
        template += "/{artists} - {title}.{output-ext}"

    # If template ends with a slash. Does not have a file name with extension
    # at the end of the template, append {artists} - {title}.{output-ext} to it
    if template.endswith("/") or template.endswith(r"\\") or template.endswith("\\\\"):
        template += "/{artists} - {title}.{output-ext}"

    # If template does not end with {output-ext}, append it to the end of the template
    if not template.endswith(".{output-ext}"):
        template += ".{output-ext}"

    formatted_string = format_query(
        song=song,
        template=template,
        santitize=True,
        file_extension=file_extension,
        short=short,
    )

    # Parse template as Path object
    file = Path(formatted_string)

    santitized_parts = []
    for part in file.parts:
        match = re.search(r"[^\.*](.*)[^\.*$]", part)
        if match and part != ".spotdl":
            santitized_parts.append(match.group(0))
        else:
            santitized_parts.append(part)

    # Join the parts of the path
    file = Path(*santitized_parts)

    # Check if the file name length is greater than 255
    if len(file.name) > 255:
        # If the file name length is greater than 255,
        # and we are already using the short version of the template,
        # fallback to default template
        if short is True:
            warnings.warn(
                f"{song.display_name}: File name is too long. Using the default template."
            )

            return create_file_name(
                song=song,
                template="/{artist} - {title}.{output-ext}",
                file_extension=file_extension,
                short=short,
            )

        # This will probably never occur, but just in case
        if short is True and template == "/{artist} - {title}.{output-ext}":
            raise RecursionError(
                f'"{song.display_name} is too long to be shortened. File a bug report on GitHub'
            )

        return create_file_name(
            song,
            template,
            file_extension,
            short=True,
        )

    return file


def parse_duration(duration: Optional[str]) -> float:
    """
    Convert string value of time (duration: "25:36:59") to a float value of seconds (92219.0)

    ### Arguments
    - duration: the string value of time

    ### Returns
    - the float value of seconds
    """

    if duration is None:
        return 0.0

    try:
        # {(1, "s"), (60, "m"), (3600, "h")}
        mapped_increments = zip([1, 60, 3600], reversed(duration.split(":")))
        seconds = sum(multiplier * int(time) for multiplier, time in mapped_increments)
        return float(seconds)

    # This usually occurs when the wrong string is mistaken for the duration
    except (ValueError, TypeError, AttributeError):
        return 0.0


def to_ms(
    string: Optional[str] = None, precision: Optional[int] = None, **kwargs
) -> float:
    """
    Convert a string to milliseconds.

    ### Arguments
    - string: the string to convert
    - precision: the number of decimals to round to
    - kwargs: the keyword args to convert

    ### Returns
    - the milliseconds

    ### Notes
    - You can either pass a string,
    - or a set of keyword args ("hour", "min", "sec", "ms") to convert.
    - If "precision" is set, the result is rounded to the number of decimals given.
    - From: https://gist.github.com/Hellowlol/5f8545e999259b4371c91ac223409209
    """

    if string:
        hour = int(string[0:2])
        minute = int(string[3:5])
        sec = int(string[6:8])
        milliseconds = int(string[10:11])
    else:
        hour = int(kwargs.get("hour", 0))
        minute = int(kwargs.get("min", 0))
        sec = int(kwargs.get("sec", 0))
        milliseconds = int(kwargs.get("ms", 0))

    result = (
        (hour * 60 * 60 * 1000) + (minute * 60 * 1000) + (sec * 1000) + milliseconds
    )

    if precision and isinstance(precision, int):
        return round(result, precision)

    return result


def restrict_filename(pathobj: Path) -> Path:
    """
    Sanitizes the filename part of a Path object. Returns modified object.

    ### Arguments
    - pathobj: the Path object to sanitize

    ### Returns
    - the modified Path object

    ### Notes
    - Based on the `sanitize_filename` function from yt-dlp
    """

    result = sanitize_filename(pathobj.name, True, False)
    result = result.replace("_-_", "-")

    if not result:
        result = "_"

    return pathobj.with_name(result)



"""
Song module that hold the Song and SongList classes.
"""

import json

from dataclasses import dataclass, asdict
from typing import Dict, Any, List, Optional

from spotdl.utils.spotify import SpotifyClient


class SongError(Exception):
    """
    Base class for all exceptions related to songs.
    """


@dataclass()
class Song:
    """
    Song class. Contains all the information about a song.
    """

    name: str
    artists: List[str]
    artist: str
    album_name: str
    album_artist: str
    genres: List[str]
    disc_number: int
    disc_count: int
    duration: int
    year: int
    date: str
    track_number: int
    tracks_count: int
    song_id: str
    cover_url: Optional[str]
    explicit: bool
    publisher: str
    url: str
    isrc: Optional[str]
    copyright_text: Optional[str]
    download_url: Optional[str] = None
    song_list: Optional["SongList"] = None
    lyrics: Optional[str] = None

    @classmethod
    def from_url(cls, url: str) -> "Song":
        """
        Creates a Song object from a URL.

        ### Arguments
        - url: The URL of the song.

        ### Returns
        - The Song object.
        """

        if "open.spotify.com" not in url or "track" not in url:
            raise SongError(f"Invalid URL: {url}")

        # query spotify for song, artist, album details
        spotify_client = SpotifyClient()

        # get track info
        raw_track_meta = spotify_client.track(url)

        if raw_track_meta is None:
            raise SongError(
                "Couldn't get metadata, check if you have passed correct track id"
            )

        # get artist info
        primary_artist_id = raw_track_meta["artists"][0]["id"]
        raw_artist_meta: Dict[str, Any] = spotify_client.artist(primary_artist_id)  # type: ignore

        # get album info
        album_id = raw_track_meta["album"]["id"]
        raw_album_meta: Dict[str, Any] = spotify_client.album(album_id)  # type: ignore

        # create song object
        return cls(
            name=raw_track_meta["name"],
            artists=[artist["name"] for artist in raw_track_meta["artists"]],
            artist=raw_track_meta["artists"][0]["name"],
            album_name=raw_album_meta["name"],
            album_artist=raw_album_meta["artists"][0]["name"],
            copyright_text=raw_album_meta["copyrights"][0]["text"]
            if raw_album_meta["copyrights"]
            else None,
            genres=raw_album_meta["genres"] + raw_artist_meta["genres"],
            disc_number=raw_track_meta["disc_number"],
            disc_count=int(raw_album_meta["tracks"]["items"][-1]["disc_number"]),
            duration=raw_track_meta["duration_ms"] / 1000,
            year=int(raw_album_meta["release_date"][:4]),
            date=raw_album_meta["release_date"],
            track_number=raw_track_meta["track_number"],
            tracks_count=raw_album_meta["total_tracks"],
            isrc=raw_track_meta.get("external_ids", {}).get("isrc"),
            song_id=raw_track_meta["id"],
            explicit=raw_track_meta["explicit"],
            publisher=raw_album_meta["label"],
            url=raw_track_meta["external_urls"]["spotify"],
            cover_url=raw_album_meta["images"][0]["url"]
            if raw_album_meta["images"]
            else None,
        )

    @classmethod
    def from_search_term(cls, search_term: str) -> "Song":
        """
        Creates a list of Song objects from a search term.

        ### Arguments
        - search_term: The search term to use.

        ### Returns
        - The Song object.
        """

        spotify_client = SpotifyClient()
        raw_search_results = spotify_client.search(search_term)

        if (
            raw_search_results is None
            or len(raw_search_results.get("tracks", {}).get("items", [])) == 0
        ):
            raise SongError("No songs matches found on spotify")

        return Song.from_url(
            "http://open.spotify.com/track/"
            + raw_search_results["tracks"]["items"][0]["id"]
        )

    @classmethod
    def from_data_dump(cls, data: str) -> "Song":
        """
        Create a Song object from a data dump.

        ### Arguments
        - data: The data dump.

        ### Returns
        - The Song object.
        """

        # Create dict from json string
        data_dict = json.loads(data)

        # Return product object
        return cls(**data_dict)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Song":
        """
        Create a Song object from a dictionary.

        ### Arguments
        - data: The dictionary.

        ### Returns
        - The Song object.
        """

        # Return product object
        return cls(**data)

    @property
    def display_name(self) -> str:
        """
        Returns a display name for the song.

        ### Returns
        - The display name.
        """

        return f"{self.artist} - {self.name}"

    @property
    def json(self) -> Dict[str, Any]:
        """
        Returns a dictionary of the song's data.

        ### Returns
        - The dictionary.
        """

        return asdict(self)


@dataclass(frozen=True)
class SongList:
    """
    SongList class. Base class for all other song lists subclasses.
    """

    name: str
    url: str
    urls: List[str]
    songs: List[Song]

    @property
    def length(self) -> int:
        """
        Get list length (number of songs).

        ### Returns
        - The list length.
        """

        return len(self.songs)

    @classmethod
    def create_basic_list(cls, url: str):
        """
        Create a basic list with only the required metadata and urls.

        ### Arguments
        - url: The url of the list.

        ### Returns
        - The SongList object.
        """

        metadata = cls.get_metadata(url)
        urls = cls.get_urls(url)

        return cls(**metadata, urls=urls, songs=[])

    @classmethod
    def from_url(cls, url: str) -> "SongList":
        """
        Initialize a SongList object from a URL.

        ### Arguments
        - url: The URL of the list.
        """

        raise NotImplementedError

    @staticmethod
    def get_urls(url: str) -> List[str]:
        """
        Get urls for all songs in url.

        ### Arguments
        - url: The URL of the list.

        ### Returns
        - The list of urls.
        """

        raise NotImplementedError

    @staticmethod
    def get_metadata(url: str) -> Dict[str, Any]:
        """
        Get metadata for list.

        ### Arguments
        - url: The URL of the list.

        ### Returns
        - The metadata.
        """

        raise NotImplementedError



"""
Lyrics providers for spotdl.
"""

from spotdl.providers.lyrics.genius import Genius
from spotdl.providers.lyrics.musixmatch import MusixMatch
from spotdl.providers.lyrics.azlyrics import AzLyrics



"""
Web module for the console.
"""

import asyncio
import logging
import os
import json
import webbrowser

from typing import Any, Dict, List, Optional, Tuple, Union
from pathlib import Path

import mimetypes


from fastapi import FastAPI, Response, WebSocket, WebSocketDisconnect, HTTPException
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel  # pylint: disable=E0611
from uvicorn import Config, Server
from starlette.types import Scope

import nest_asyncio

from spotdl.download.downloader import Downloader, DownloaderError
from spotdl.download.progress_handler import NAME_TO_LEVEL, ProgressHandler, SongTracker
from spotdl.types.song import Song
from spotdl.utils.github import download_github_dir
from spotdl.utils.search import parse_query
from spotdl.utils.search import get_search_results
from spotdl.utils.config import get_spotdl_path


ALLOWED_ORIGINS = [
    "http://localhost:8800",
    "https://127.0.0.1:8800",
    "http://localhost:3000",
    "http://localhost:8080",
    "*",
]


class App:
    """
    App class that holds the application state.
    """

    server: Any
    downloader: Downloader
    settings: Dict[str, Any]
    loop: asyncio.AbstractEventLoop


class SongModel(BaseModel):
    """
    A song object used for types and validation.
    We can't use the Song class directly because FastAPI doesn't support dataclasses.
    """

    name: str
    artists: List[str]
    artist: str
    album_name: str
    album_artist: str
    genres: List[str]
    disc_number: int
    disc_count: int
    copyright: str
    duration: int
    year: int
    date: str
    track_number: int
    tracks_count: int
    isrc: str
    song_id: str
    cover_url: Optional[str]
    explicit: bool
    publisher: str
    url: str
    download_url: Optional[str] = None


class SettingsModel(BaseModel):
    """
    A settings object used for types and validation.
    """

    log_level: Optional[str]
    cache_path: Optional[str]
    audio_providers: Optional[List[str]]
    lyrics_providers: Optional[List[str]]
    ffmpeg: Optional[str]
    variable_bitrate: Optional[int]
    constant_bitrate: Optional[int]
    ffmpeg_args: Optional[List[str]]
    format: Optional[str]
    save_file: Optional[str]
    m3u: Optional[str]
    output: Optional[str]
    overwrite: Optional[str]
    client_id: Optional[str]
    client_secret: Optional[str]
    user_auth: Optional[bool]
    threads: Optional[int]


app = App()
app.server = FastAPI()
app.server.add_middleware(
    CORSMiddleware,
    allow_origins=ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

nest_asyncio.apply()


class WSProgressHandler:
    """
    Handles all the WebSocket connections.
    """

    instances: List["WSProgressHandler"] = []

    def __init__(self, websocket: WebSocket, client_id: str):
        """
        Initialize the WebSocket handler.

        ### Arguments
        - websocket: The WebSocket instance.
        - client_id: The client's ID.
        """

        self.client_id = client_id
        self.websocket = websocket

    async def connect(self):
        """
        Called when a new client connects to the websocket.
        """

        connection = {"client_id": self.client_id, "websocket": self.websocket}
        logging.info("Connecting WebSocket: %s", connection)
        await self.websocket.accept()
        WSProgressHandler.instances.append(self)

    @classmethod
    def get(cls, client_id: str):
        """
        Get a WSProgressHandler instance by client_id.

        ### Arguments
        - client_id: The client's ID.
        """

        try:
            instance = next(
                inst for inst in cls.instances if inst.client_id == client_id
            )
            return instance
        except StopIteration:
            logging.warning(
                "Error while accessing websocket instance. Websocket not created"
            )
            return None

    async def send_update(self, message: str):
        """
        Send an update to the client.

        ### Arguments
        - message: The message to send.
        """

        logging.debug("Sending %s: %s", self.client_id, message)
        await self.websocket.send_text(message)

    def update(self, progress_handler_instance: SongTracker, message: str):
        """Callback function from ProgressHandler's SongTracker, called on every update

        ### Arguments
        - progress_handler_instance: The ProgressHandler instance.
        - message: The message to send.
        """
        update_message = {
            "song": progress_handler_instance.song.json,
            "progress": progress_handler_instance.progress,
            "message": message,
        }
        asyncio.run(self.send_update(json.dumps(update_message)))


def fix_mime_types():
    """Fix incorrect entries in the `mimetypes` registry.
    On Windows, the Python standard library's `mimetypes` reads in
    mappings from file extension to MIME type from the Windows
    registry. Other applications can and do write incorrect values
    to this registry, which causes `mimetypes.guess_type` to return
    incorrect values, which causes spotDL to fail to render on
    the frontend.
    This method hard-codes the correct mappings for certain MIME
    types that are known to be either used by TensorBoard or
    problematic in general.
    """
    # Known to be problematic when Visual Studio is installed:
    # <https://github.com/tensorflow/tensorboard/issues/3120>
    # https://github.com/spotDL/spotify-downloader/issues/1540
    mimetypes.add_type("application/javascript", ".js")
    # Not known to be problematic, but used by spotDL:
    mimetypes.add_type("text/css", ".css")
    mimetypes.add_type("image/svg+xml", ".svg")
    mimetypes.add_type("text/html", ".html")


@app.server.websocket("/api/ws")
async def websocket_endpoint(websocket: WebSocket, client_id: str):
    """
    Websocket endpoint.

    ### Arguments
    - websocket: The WebSocket instance.
    - client_id: The client's ID.
    """

    await WSProgressHandler(websocket, client_id).connect()

    try:
        while True:
            data = await websocket.receive_text()
            logging.debug("Client %s said: %s", client_id, data)
    except WebSocketDisconnect:
        logging.info("Disconnecting WebSocket: %s", client_id)


@app.server.get("/api/song/search")
def song_from_search(query: str) -> Song:
    """
    Search for a song on spotify using search query.

    ### Arguments
    - query: The search query.

    ### Returns
    - returns the first result as a Song object.
    """

    return Song.from_search_term(query)


@app.server.get("/api/song/url")
def song_from_url(url: str) -> Song:
    """
    Search for a song on spotify using url.

    ### Arguments
    - url: The url to search.

    ### Returns
    - returns the first result as a Song object.
    """

    return Song.from_url(url)


@app.server.post("/api/songs/query")
def query_search(query: List[str]) -> List[Song]:
    """
    Parse a list of search queries.

    ### Arguments
    - query: The query to parse.

    ### Returns
    - returns a list of Song objects.
    """

    return parse_query(query)


@app.server.get("/api/songs/search")
def search_search(query: str) -> List[Song]:
    """
    Parse search term and return list of Song objects.

    ### Arguments
    - query: The query to parse.

    ### Returns
    - returns a list of Song objects.
    """

    # return parse_query([query], app.downloader.threads)
    return get_search_results(query)


@app.server.post("/api/downloader/change_output")
def change_output(output: str) -> bool:
    """
    Change output folder

    ### Arguments
    - output: The output folder.

    ### Returns
    - returns True if the output folder was changed.
    """

    app.downloader.output = output

    return True


@app.server.post("/api/downloader/download/search")
async def download_search(
    query: str, return_file: bool = False
) -> Union[Tuple[Song, Optional[Path]], FileResponse]:
    """
    Search for song and download the first result.

    ### Arguments
    - query: The query to search.
    - return_file: If True, return the file instead of the song.

    ### Returns
    - returns a Tuple of Song and Path if return_file is True.
    - returns a FileResponse if return_file is False.
    """

    song, path = await app.downloader.pool_download(Song.from_search_term(query))

    if return_file is True:
        if path is None:
            raise ValueError("No file found")

        return FileResponse(path)

    return song, path


@app.server.post("/api/download/objects")
async def download_objects(
    song: SongModel, return_file: bool = False
) -> Union[Tuple[Song, Optional[Path]], FileResponse]:
    """
    Download songs using Song objects.

    ### Arguments
    - song: The Song object.
    - return_file: If True, return the file instead of the song.

    ### Returns
    - returns a Tuple of Song and Path if return_file is True.
    - returns a FileResponse if return_file is False.
    """

    song_obj, path = await app.downloader.pool_download(Song(**song.dict()))

    if return_file is True:
        if path is None:
            raise ValueError("No file found")

        return FileResponse(path)

    return song_obj, path


@app.server.post("/api/download/url")
async def download_url(url: str, client_id: str) -> Optional[str]:
    """
    Download songs using Song url.

    ### Arguments
    - url: The url to download.
    - client_id: The client's ID.

    ### Returns
    - returns the file path if the song was downloaded.
    """

    app.downloader.output = str(
        (get_spotdl_path() / f"web/sessions/{client_id}").absolute()
    )

    # Initiate realtime updates if websocket from client is connected
    ws_instance = WSProgressHandler.get(client_id)
    if ws_instance is not None:
        app.downloader.progress_handler = ProgressHandler(
            NAME_TO_LEVEL[app.settings["log_level"]],
            simple_tui=True,
            update_callback=ws_instance.update,
        )

    try:
        # Fetch song metadata
        song = Song.from_url(url)

        # Download Song
        _, path = await app.downloader.pool_download(song)

        if path is None:
            exc = DownloaderError(f"Failure downloading {song.name}")
            logging.warning("Error downloading! %s", exc)
            raise HTTPException(status_code=500, detail=f"Error downloading: {exc}")

        # Strip Filename
        filename = os.path.basename(path)

        return filename

    except Exception as exception:
        logging.warning("Error downloading! %s", exception)
        raise HTTPException(
            status_code=500, detail=f"Error downloading: {exception}"
        ) from exception


@app.server.get("/api/download/file")
async def download_file(file: str, client_id: str) -> FileResponse:
    """
    Download file using path.

    ### Arguments
    - file: The file path.
    - client_id: The client's ID.

    ### Returns
    - returns the file response, filename specified to return as attachment.
    """

    return FileResponse(
        str((get_spotdl_path() / f"web/sessions/{client_id}/{file}").absolute()),
        filename=file,
    )


@app.server.post("/api/download/multiple_search")
def download_multiple_search(query: List[str]) -> List[Tuple[Song, Optional[Path]]]:
    """
    Search for song and download the first result.

    ### Arguments
    - query: The query to search.

    ### Returns
    - returns a list of Tuple of Song and Path if the song was downloaded.
    """

    return app.downloader.download_multiple_songs(parse_query(query))


@app.server.post("/api/download/multiple_objects")
def download_multiple_objects(
    songs: List[SongModel],
) -> List[Tuple[Song, Optional[Path]]]:
    """
    Download songs using Song objects.

    ### Arguments
    - songs: The list of Song objects.

    ### Returns
    - returns a list of Tuple of Song and Path if the song was downloaded.
    """

    return app.downloader.download_multiple_songs(
        [Song.from_dict(song.dict()) for song in songs]
    )


@app.server.get("/api/settings")
def get_settings() -> SettingsModel:
    """
    Return the settings object.

    ### Returns
    - returns the settings object.
    """

    return SettingsModel(**app.settings)


@app.server.post("/api/settings/update")
def change_settings(settings: SettingsModel) -> bool:
    """
    Change downloader settings by re-initializing the downloader.

    ### Arguments
    - settings: The settings to change.

    ### Returns
    - returns True if the settings were changed.
    """

    settings_dict = settings.dict()

    # Create shallow copy of settings
    settings_cpy = app.settings.copy()

    # Update settings with new settings that are not None
    settings_cpy.update({k: v for k, v in settings_dict.items() if v is not None})

    logging.debug("Applying settings: %s", {settings_cpy})

    # Re-initialize downloader
    app.downloader = Downloader(
        audio_providers=settings_cpy["audio_providers"],
        lyrics_providers=settings_cpy["lyrics_providers"],
        ffmpeg=settings_cpy["ffmpeg"],
        bitrate=settings_cpy["bitrate"],
        ffmpeg_args=settings_cpy["ffmpeg_args"],
        output_format=settings_cpy["format"],
        save_file=settings_cpy["save_file"],
        threads=settings_cpy["threads"],
        output=settings_cpy["output"],
        overwrite=settings_cpy["overwrite"],
        log_level="CRITICAL",
        simple_tui=True,
        loop=app.loop,
    )

    return True


class SPAStaticFiles(StaticFiles):
    """
    Override the static files to serve the index.html and other assets.
    """

    async def get_response(self, path: str, scope: Scope) -> Response:
        """
        Serve static files from the SPA.

        ### Arguments
        - path: The path to the file.
        - scope: The scope of the request.

        ### Returns
        - returns the response.
        """

        response = await super().get_response(path, scope)
        if response.status_code == 404:
            response = await super().get_response(".", scope)

        return response


def web(settings: Dict[str, Any]):
    """
    Run the web server.

    ### Arguments
    - settings: Settings dictionary, based on the `SettingsModel` class.
    """

    fix_mime_types()

    web_app_dir = str(get_spotdl_path().absolute())

    print("Updating web app")

    # Get web client from CDN (github for now)
    download_github_dir(
        "https://github.com/spotdl/web-ui/tree/master/dist", output_dir=web_app_dir
    )

    # Serve web client SPA
    app.server.mount(
        "/", SPAStaticFiles(directory=web_app_dir + "/dist", html=True), name="static"
    )

    loop = asyncio.new_event_loop()

    app.loop = loop
    app.settings = settings
    app.downloader = Downloader(
        audio_providers=settings["audio_providers"],
        lyrics_providers=settings["lyrics_providers"],
        ffmpeg=settings["ffmpeg"],
        bitrate=settings["bitrate"],
        ffmpeg_args=settings["ffmpeg_args"],
        output_format=settings["format"],
        save_file=settings["save_file"],
        threads=settings["threads"],
        output=settings["output"],
        overwrite=settings["overwrite"],
        log_level=settings["log_level"],
        simple_tui=True,
        loop=loop,
    )

    config = Config(app=app.server, port=8800, workers=1, loop=loop)  # type: ignore

    server = Server(config)

    webbrowser.open("http://localhost:8800")

    loop.run_until_complete(server.serve())

    app.downloader.progress_handler.close()
